#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <libpq-fe.h>
#include <pcre.h>

int log2db(PGconn *conn, const char *logname, const char *logitems, char *logvalues);
int read_and_log(const char *infile, const char *outfile,const char *logname, const char *logitems, int logtype);
int usage(const char *filename,int versioninfo);
void writelog(const char *outfile,char *output);
void pglogger(int priority,const char *message);
char *parselog(char *logline, int logtype);

int daemonized;

int main(int argc, char **argv) {
    const char apache_logitems[] = "clientaddress,rfc1413,username,\"localtime\",httprequest,statuscode,bytessenttoclient,referer,clientsoftware,timespent,canonname";
    const char squid_logitems[] = "clientaddress,rfc1413,username,\"localtime\",httprequest,statuscode,bytessenttoclient,cacheinfo";
    char *infile = NULL;
    char *outfile = NULL;
    int c,logtype = 0;
    struct stat buf;
    
    opterr = 0;
    daemonized = 0;

    while(1) {
	c = getopt(argc,argv,"vdi:o:t:");

	if(c == -1)
	    break;

	switch(c) {
	    case '?':
	    case ':':
		usage(argv[0],0);
		exit(1);
	    case 'v':
		usage("",1);
		exit(1);
	    case 'd':
		daemonized = 1;
		break;
	    case 'i':
		if((infile = strdup(optarg)) == NULL) {
		    fprintf(stderr,"out of memory allocating buffer for infile string");
		    exit(1);
		}
		break;
	    case 'o':
		if((outfile = strdup(optarg)) == NULL) {
		    fprintf(stderr,"out of memory allocating buffer for outfile string");
		    exit(1);
		}
		break;
	    case 't':
		if(!strcmp(optarg,"apache"))
		    logtype = 1;
		else if(!strcmp(optarg,"squid"))
		    logtype = 2;
		else {
		    usage(argv[0],0);
		    exit(1);
		}
		break;
	    default:
		usage(argv[0],0);
		exit(1);
	}
    }
    
    if((optind < argc) || !logtype || !infile || !outfile) {
	usage(argv[0],0);
	printf("options -t apache|squid, -i filename, and -o filename are compulsory\n");
	exit(1);
    }
    
    if(daemonized && (!strcmp(infile,"-") || !strcmp(infile,"/dev/stdin"))) {
	usage(argv[0],0);
	fprintf(stderr,"-d requires -i be different from stdin\n");
	exit(1);
    }

    if(!strcmp(infile,"/dev/stdout")) {
	usage(argv[0],0);
	fprintf(stderr, "cannot read from stdout, did you mean -i -?\n");
	exit(1);
    }
    if(!strcmp(outfile,"/dev/stdin")) {
	usage(argv[0],0);
	fprintf(stderr,"cannot write to stdin, did you mean -o -?\n");
	exit(1);
    }
    
    if(stat(infile,&buf) && strcmp(infile,"-")) {
	fprintf(stderr,"%s does not exist\n",infile);
	exit(1);
    } else {
	if(daemonized && !S_ISFIFO(buf.st_mode)) {
	    fprintf(stderr,"WARNING: -d assumes %s be a fifo\n",infile);
	    fprintf(stderr,"(otherwise the file would be logged repeatedly)\n");
	    fprintf(stderr,"Use mkfifo %s to create one\n",infile);
	    exit(1);
	}
    }

    if(daemonized) daemon(1,0);

    do {
	switch(logtype) {
	    case 1:
		read_and_log(infile, outfile,"apachelog", apache_logitems, logtype);
		break;
	    case 2:
		read_and_log(infile, outfile,"squidlog", squid_logitems, logtype);
		break;
	}
    } while(daemonized);
    exit(0);
}

void pglogger(int priority,const char *message) {
    if(daemonized)
	openlog("pglogger",0,LOG_DAEMON);
    else
	openlog("pglogger",0,LOG_USER);
    syslog(priority,message);
    closelog();
}

int usage(const char *filename,int versioninfo) {
    printf("pglogger version 2.0 - (c) Copyright 2003 GeekNET\n");
    if(versioninfo) return(0);
    printf("Usage: %s -[vd]iot\n",filename);
    printf("\t-v\t\tPrint version info and exit\n");
    printf("\t-d\t\tDaemonize, requires -i filename, which must be a FIFO\n");
    printf("\t-i filename\tRead from filename\n");
    printf("\t-i -\t\tRead from stdin\n");
    printf("\t-o filename\twrite to filename\n");
    printf("\t-o -\t\twrite to stdout\n");
    printf("\t-t apache|squid\tLogformat, logfile must be in either\n");
    printf("\t\t\tapache Common Logfile Format, including timespent\n");
    printf("\t\t\tand hostname, or in Squid's apache emulation format.\n\n");
    return(-1);
}

int log2db(PGconn *conn, const char *logname, const char *logitems, char *logvalues) {
    char *buffer;
    int retry_count = 0;
    PGresult *res = NULL;
    const char sqlstring[] = "INSERT INTO %s (%s) VALUES (%s)";
    
    if((buffer = (char *)malloc(snprintf(buffer,0,sqlstring,logname,logitems,logvalues)+1)) == NULL) {
	return(-1);
    }
    sprintf(buffer,sqlstring,logname,logitems,logvalues);
    res = PQexec(conn,buffer);
    while((retry_count < 3) && (PQstatus(conn) == CONNECTION_BAD)) {
	PQclear(res);
	PQreset(conn);
	res = PQexec(conn,buffer);
	retry_count++;
    }
    free(buffer);
    if((PQstatus(conn) == CONNECTION_BAD) || (PQresultStatus(res) != PGRES_COMMAND_OK)) {
	PQclear(res);
	return(-1);
    }
    PQclear(res);
    return(0);
}

char *parselog(char *logline, int logtype) {
    int offsets[36];
    int size_offsets = sizeof(offsets)/sizeof(int);
    int erroroffset,rc,count;
    long int statuscode,bytessenttoclient,timespent;
    const char *error, **logvalues;
    pcre *re = NULL;

    const char apachepreg[] = "^(\\S+) (\\S+) (\\S+) \\[(.+)\\] \"(.+)\" (\\S+) (\\S+) \"(.+)\" \"(.+)\" (\\S+) (\\S+)$";
    const char squidpreg[] = "^(\\S+) (\\S+) (\\S+) \\[(.+)\\] \"(.+)\" (\\S+) (\\S+) (\\S+)$";
    const char apacheformat [] = "'%.255s','%.32s','%.32s','%s','%.255s',%ld,%ld,'%.255s','%.255s',%ld,'%.255s'";
    const char squidformat[] = "'%.255s','%.32s','%.32s','%s','%.255s',%ld,%ld,'%.255s'";

    char *temp;
    char *username;
    char *httprequest;
    char *referer;
    char *clientsoftware;
    char *canonname;

    char *newlogline = strdup(logline);

    if(newlogline == NULL) {
	return(NULL);
    }
    if(logtype == 1) {
	if((re = pcre_compile((char *)apachepreg, PCRE_UNGREEDY, &error, &erroroffset, NULL)) != NULL) {
	    if((count = pcre_exec(re, NULL, newlogline, strlen(newlogline),0, 0, offsets, size_offsets)) >= 0) {
		rc = pcre_get_substring_list(newlogline, offsets, count,&logvalues);
		if(rc == PCRE_ERROR_NOMEMORY) {
		    free(newlogline);
		    return(NULL);
		}
	    } else {
		pcre_free(re);
		pglogger(LOG_ERR,"Parser found no matches");
		free(newlogline);
		return(NULL);
	    }
	} else {
	    free(newlogline);
	    pglogger(LOG_ERR,"Regular expression did not compile");
	    return(NULL);
	}
    } else if(logtype == 2) {
	if((re = pcre_compile((char *)squidpreg, PCRE_UNGREEDY, &error, &erroroffset, NULL)) != NULL) {
	    if((count = pcre_exec(re, NULL, newlogline, strlen(newlogline),0, 0, offsets, size_offsets)) >= 0) {
		rc = pcre_get_substring_list(newlogline, offsets, count,&logvalues);
		if(rc == PCRE_ERROR_NOMEMORY) {
		    free(newlogline);
		    return(NULL);
		}
	    } else {
		pcre_free(re);
		pglogger(LOG_ERR,"Parser found no matches");
		free(newlogline);
		return(NULL);
	    }
	} else {
	    free(newlogline);
	    pglogger(LOG_ERR,"Regular expression did not compile");
	    return(NULL);
	}
    } else {
	free(newlogline);
	pglogger(LOG_ERR,"unknown logtype");
	return(NULL);
    }
    pcre_free(re);
    
    if(count == 12) {
	timespent = strtol(logvalues[10],NULL,10);
	statuscode = strtol(logvalues[6],NULL,10);
	bytessenttoclient = strtol(logvalues[7],NULL,10);
	
	if((username = malloc((strlen(logvalues[3])*2)+1)) != NULL)
	    PQescapeString(username,logvalues[3],strlen(logvalues[3]));
	if((httprequest = malloc((strlen(logvalues[5])*2)+1)) != NULL)
	    PQescapeString(httprequest,logvalues[5],strlen(logvalues[5]));
	if((referer = malloc((strlen(logvalues[8])*2)+1)) != NULL)
	    PQescapeString(referer,logvalues[8],strlen(logvalues[8]));
	if((clientsoftware = malloc((strlen(logvalues[9])*2)+1)) != NULL)
	    PQescapeString(clientsoftware,logvalues[9],strlen(logvalues[9]));
	if((canonname = malloc((strlen(logvalues[11])*2)+1)) != NULL)
	    PQescapeString(canonname,logvalues[11],strlen(logvalues[11]));
	
	if((temp = (char *)realloc(newlogline,snprintf(newlogline,0,apacheformat,
	    logvalues[1],logvalues[2],username,logvalues[4],
	    httprequest,statuscode,bytessenttoclient,referer,
	    clientsoftware,timespent,canonname)+1)) == NULL) {
	    free(username);
	    free(httprequest);
	    free(referer);
	    free(clientsoftware);
	    free(canonname);
	    free(newlogline);
	    pcre_free_substring_list(logvalues);
	    return(NULL);
	} else newlogline = temp;
	sprintf(newlogline,apacheformat,
	    logvalues[1],logvalues[2],username,logvalues[4],
	    httprequest,statuscode,bytessenttoclient,referer,
	    clientsoftware,timespent,canonname);
	free(username);
	free(httprequest);
	free(referer);
	free(clientsoftware);
	free(canonname);
    } else if(count == 9) {
	if((username = malloc((strlen(logvalues[3])*2)+1)) != NULL)
	    PQescapeString(username,logvalues[3],strlen(logvalues[3]));
	if((httprequest = malloc((strlen(logvalues[5])*2)+1)) != NULL)
	    PQescapeString(httprequest,logvalues[5],strlen(logvalues[5]));
	statuscode = strtol(logvalues[6],NULL,10);
	bytessenttoclient = strtol(logvalues[7],NULL,10);
	if((temp = (char *)realloc(newlogline,snprintf(newlogline,0,squidformat,
	    logvalues[1],logvalues[2],username,logvalues[4],
	    httprequest,statuscode,bytessenttoclient,logvalues[8])+1)) == NULL) {
	    free(newlogline);
	    pcre_free_substring_list(logvalues);
	    free(username);
	    free(httprequest);
	    return(NULL);
	} else newlogline = temp;
	sprintf(newlogline,squidformat,
	    logvalues[1],logvalues[2],username,logvalues[4],
	    httprequest,statuscode,bytessenttoclient,logvalues[8]);
	free(username);
	free(httprequest);
    } else {
	pcre_free_substring_list(logvalues);
	pglogger(LOG_ERR,"Unmatched number of results from parser");
	return(NULL);
    }
    pcre_free_substring_list(logvalues);
    return(newlogline);
}

int read_and_log(const char *infile, const char *outfile, const char *logname, const char *logitems, int logtype) {
    char *buffer, *output, *logline, *temp;
    int buflen,count = 0;
    FILE *infilep,*outfilep;
    PGconn *conn;
    
    if(!strcmp(infile,"-"))
	infilep = stdin;
    else
	infilep = fopen(infile,"r");
    if(infilep != NULL) {
	conn = PQconnectdb("host=log-db dbname=logs");
    	if (PQstatus(conn) != CONNECTION_BAD) {
	    do {
		while(!feof(infilep)) {
		    if((buffer = (char *)malloc(65)) == NULL) {
			PQfinish(conn);
			return(0);
		    }
		    memset(buffer,0,65);
		    fgets(buffer, 65, infilep);
		    buflen = 64;
		    if(strlen(buffer) == 0) {
			free(buffer);
			if(feof(infilep))
			    break;
			else
			    continue;
		    }
		    while((strchr(buffer,'\n') == NULL) && !feof(infilep)) {
			if((temp = (char *)realloc(buffer,buflen + 65)) == NULL) {
			    free(buffer);
			    PQfinish(conn);
			    return(0);
			} else buffer = temp;
			fgets(buffer + buflen, 65, infilep);
    			buflen += 64;
		    }
		    if((output = (char *)strndup(buffer, strlen(buffer)-1)) == NULL) {
			free(buffer);
			PQfinish(conn);
			return(0);
		    }
		    free(buffer);
		    if((logline = parselog(output,logtype)) != NULL) {
			if(log2db(conn, logname, logitems, logline)) {
			    pglogger(LOG_ERR,"logging failed, writing to outfile instead");
			    pglogger(LOG_ERR,logline);
			    writelog(outfile,output);
			}
		    } else {
			pglogger(LOG_ERR,"parsing failed, writing to outfile instead");
			writelog(outfile,output);
		    }
		    free(logline);
		    free(output);
		}
	    } while(daemonized);
	} else {
	    pglogger(LOG_CRIT,"unable to connect to database");
	}
	PQfinish(conn);
    } else {
	pglogger(LOG_CRIT,"unable to open infile for reading");
    }
    return(1);
}

void writelog(const char *outfile,char *output) {
    FILE *outfilep;

    if(!strcmp(outfile,"-"))
	outfilep = stdout;
    else
	outfilep = fopen(outfile,"a+");
    if(outfilep != NULL) {
	fprintf(outfilep,"%s\n",output);
	if(strcmp(outfile,"-"))
	    fclose(outfilep);
    } else {
	pglogger(LOG_CRIT,"unable to open outfile for writing");
    }
}
